// Crimson Fields -- a game of tactical warfare
// Copyright (C) 2000-2002 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

////////////////////////////////////////////////////////////////////////
// lset.cpp -- level set handling
//
// History:
//  23-02-2002 - created
////////////////////////////////////////////////////////////////////////

#include <string>
using namespace std;

#include "SDL_endian.h"
#include "lset.h"
#include "fileio.h"


////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::LevelSet
// DESCRIPTION: Initialize a new level set.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

LevelSet::LevelSet( void ) {
  num_tt = num_ut = num_sfx = 0;
  tt = NULL;
  ut = NULL;
  sfx = NULL;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::~LevelSet
// DESCRIPTION: Destroy the currently loaded level set.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

LevelSet::~LevelSet( void ) {
  delete [] tt;
  delete [] ut;
  if ( sfx ) {
    for ( int i = 0; i < num_sfx; ++i ) delete sfx[i];
    delete [] sfx;
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::GetTerrainInfo
// DESCRIPTION: Retrieve information about a terrain type by its ID.
// PARAMETERS : ttid - terrain type ID
// RETURNS    : pointer to terrain info or NULL if ID is invalid
//
// HISTORY
////////////////////////////////////////////////////////////////////////

const TerrainType *LevelSet::GetTerrainInfo( unsigned short ttid ) const {
  if ( ttid < num_tt ) return &tt[ttid];
  return NULL;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::GetUnitInfo
// DESCRIPTION: Retrieve information about a unit type by its ID.
// PARAMETERS : utid - unit type ID
// RETURNS    : pointer to unit info or NULL if ID is invalid
//
// HISTORY
////////////////////////////////////////////////////////////////////////

const UnitType *LevelSet::GetUnitInfo( unsigned short utid ) const {
  if ( utid < num_ut ) return &ut[utid];
  return NULL;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::GetSound
// DESCRIPTION: Get a sound effect by its ID.
// PARAMETERS : sfxid - effect ID
// RETURNS    : pointer to effect or NULL if ID is invalid
//
// HISTORY
////////////////////////////////////////////////////////////////////////

SoundEffect *LevelSet::GetSound( unsigned short sfxid ) const {
  if ( sfxid < num_sfx ) return sfx[sfxid];
  return NULL;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::Load
// DESCRIPTION: Load a level set from a file.
// PARAMETERS : setname - name of the requested level set. The actual
//                        filename is computed by prefixing it with the
//                        data directory path and adding a .set suffix.
// RETURNS    : 0 on success, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int LevelSet::Load( const char *setname ) {
  int rc = -1;
  string fname( get_data_dir() );

  fname.append( setname );
  fname.append( ".set" );

  SDL_RWops *file = SDL_RWFromFile( fname.c_str(), "rb" );
  if ( !file ) return -1;

  if ( SDL_ReadLE32( file ) == FID_SET ) {
    unsigned char sounds;

    num_ut = SDL_ReadLE16( file );
    num_tt = SDL_ReadLE16( file );
    SDL_RWread( file, &sounds, 1, 1 );
    num_sfx = sounds;

    if ( !LoadSfx( file ) && !LoadUnitTypes( file ) &&
         !LoadTerrainTypes( file ) && !LoadGfx( file ) ) {
      name.assign( setname );
      rc = 0;
    }
  }

  SDL_RWclose( file );
  return rc;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::LoadTerrainTypes
// DESCRIPTION: Load terrain type definitions from a level set file.
// PARAMETERS : file - SDL_RWops descriptor of the level set file
// RETURNS    : 0 on success, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int LevelSet::LoadTerrainTypes( SDL_RWops *file ) {
  tt = new TerrainType [num_tt];
  if ( !tt ) return -1;

  for ( int i = 0; i < num_tt; ++i ) {
    if ( tt[i].Load( file ) ) {
      delete [] tt;
      tt = NULL;
      num_tt = 0;
      return -1;
    }
  }
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::LoadUnitTypes
// DESCRIPTION: Load unit type definitions from a level set file.
// PARAMETERS : file - SDL_RWops descriptor of the level set file
// RETURNS    : 0 on success, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int LevelSet::LoadUnitTypes( SDL_RWops *file ) {
  ut = new UnitType [num_ut];
  if ( !ut ) return -1;

  for ( int i = 0; i < num_ut; ++i ) {
    if ( ut[i].Load( file, this ) ) {
      delete [] ut;
      ut = NULL;
      num_ut = 0;
      return -1;
    }
  }
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::LoadSfx
// DESCRIPTION: Load unit sound effects from a level set file.
// PARAMETERS : file - SDL_RWops descriptor of the level set file
// RETURNS    : 0 on success, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int LevelSet::LoadSfx( SDL_RWops *file ) {

  if ( num_sfx ) {
    string sfxd( get_sounds_dir() );
    char sfxname[256], len;

    sfx = new SoundEffect * [num_sfx];
    if ( !sfx ) return -1;

    for ( int i = 0; i < num_sfx; ++i ) {
      SDL_RWread( file, &len, 1, 1 );
      SDL_RWread( file, sfxname, 1, len );
      sfxname[len] = '\0';
      sfx[i] = new SoundEffect( (sfxd + sfxname).c_str() );
    }
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : LevelSet::LoadGfx
// DESCRIPTION: Load the graphics set to use for the map from a file.
// PARAMETERS : file - SDL_RWops descriptor of a data file
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int LevelSet::LoadGfx( SDL_RWops *file ) {
  if ( tiles.LoadImageData( file ) ) return -1;
  tiles.DisplayFormat();
  return 0;
}

