// Crimson Fields -- a game of tactical warfare
// Copyright (C) 2000-2002 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

////////////////////////////////////////////////////////////////////////
// map.cpp
//
// History:
//  06-12-2000 - created
//  05-03-2001 - FindPath() now starts from source; added deviation
//             - minor code clean-up
//  08-03-2001 - preliminary fix for path blocking in FindPath()
//  12-03-2001 - path blocking fix
//  17-03-2002 - pathfinding moved to separate class
////////////////////////////////////////////////////////////////////////

#include "SDL_endian.h"

#include "map.h"
#include "game.h"
#include "history.h"
#include "globals.h"

////////////////////////////////////////////////////////////////////////
// NAME       : TerrainType::Load
// DESCRIPTION: Load a terrain type definition from a file.
// PARAMETERS : file - SDL_RWops file descriptor
// RETURNS    : 0 on success, -1 on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int TerrainType::Load( SDL_RWops *file ) {
  tt_type = SDL_ReadLE16( file );
  tt_image = SDL_ReadLE16( file );
  SDL_RWread( file, &tt_att_mod, 1, 3 );
  tt_color = SDL_ReadLE32( file );
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::Map
// DESCRIPTION: Create a new map instance and initilize data.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Map::Map( void ) {
  m_data = NULL;
  m_objects = NULL;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::~Map
// DESCRIPTION: Free memory allocated by the map structure.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Map::~Map( void ) {
  delete [] m_data;
  delete [] m_objects;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::Load
// DESCRIPTION: Initialize the map structure.
// PARAMETERS : file - descriptor of an opened data file from which to
//                     read the map data
// RETURNS    : 0 on success, non-zero on error
//
// HISTORY
//   21-10-2001 - fixed a bug introduced in the last release which led
//                to an arbitrary map size (and usually a crash)
////////////////////////////////////////////////////////////////////////

int Map::Load( SDL_RWops *file ) {
  m_w = SDL_ReadLE16( file );
  m_h = SDL_ReadLE16( file );

  unsigned short size = m_w * m_h;
  m_data = new short [size];
  m_objects = new MapObject * [size];

  // terrain types are loaded from a level set 

  for ( int i = 0; i < size; ++i ) {
    m_data[i] = SDL_ReadLE16( file );
    m_objects[i] = NULL;
  }

  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::Save
// DESCRIPTION: Save the current map status to a file. Only the terrain
//              type information is saved for each hex. Other data
//              (units, buildings...) must be stored separately.
// PARAMETERS : file - descriptor of the save file
// RETURNS    : 0 on success, non-zero on error
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int Map::Save( SDL_RWops *file ) const {
  SDL_WriteLE16( file, m_w );
  SDL_WriteLE16( file, m_h );

  unsigned short size = m_w * m_h;
  for ( int i = 0; i < size; ++i )
    if ( SDL_WriteLE16( file, m_data[i] ) != 1 ) return -1;
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::GetNeighbors
// DESCRIPTION: Get the hex coordinates of all hexes surrounding a hex.
// PARAMETERS : x      - horizontal position of hex
//              y      - vertical position of hex
//              parray - pointer to an array of at least six Points to
//                       hold the neighbors' coordinates. Each neighbor
//                       will be put in the slot corresponding to the
//                       direction the hex is in (e.g. the hex north of
//                       the source hex will reside in parray[NORTH]).
//                       If there is no valid neighbor (because the map
//                       ends here), the respective Points will contain
//                       {-1, -1}
// RETURNS    : number of adjacent hexes found
//
// HISTORY
////////////////////////////////////////////////////////////////////////

short Map::GetNeighbors( short x, short y, Point *parray ) const {
  short num = 0;
  for ( int i = NORTH; i <= NORTHWEST; ++i ) {
    if ( !Dir2XY( x, y, (Direction)i, parray[i] ) ) ++num;
    else parray[i].x = parray[i].y = -1;
  }
  return num;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::Dir2XY
// DESCRIPTION: Get the coordinates of the hex we'll be on when we move
//              in a given direction from another hex.
// PARAMETERS : x    - hex horizontal position
//              y    - hex vertical position
//              dir  - direction to move in
//              dest - pointer to a Point to hold the coordinates of the
//                     destination hex
// RETURNS    : 0 on success, -1 on error (the contents of the Point can
//              not be relied on in this case)
//
// HISTORY
////////////////////////////////////////////////////////////////////////

int Map::Dir2XY( short x, short y, Direction dir, Point &dest ) const {
  switch ( dir ) {
    case NORTH:
      --y;
      break;
    case SOUTH:
      ++y;
      break;
    case NORTHEAST:
      if ( x & 1 ) --y;    // fall through...
    case EAST:
      ++x;
      break;
    case SOUTHEAST:
      if ( !(x & 1) ) ++y;
      ++x;
      break;
    case SOUTHWEST:
      if ( !(x & 1) ) ++y; // fall through...
    case WEST:
      --x;
      break;
    case NORTHWEST:
      if ( x & 1 ) --y;
      --x;
  }

  if ( !Contains( x, y ) ) return -1;

  dest.x = x;
  dest.y = y;
  return 0;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::Contains
// DESCRIPTION: Check whether a hex with the given coordinates exists
//              on the map.
// PARAMETERS : x - hex horizontal position
//              y - hex vertical position
// RETURNS    : true if the coordinates denote a valid hex, false
//              otherwise
//
// HISTORY
////////////////////////////////////////////////////////////////////////

bool Map::Contains( short x, short y ) const {
  return( (x >= 0) && (x < m_w) && (y >= 0) && (y < m_h) );
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::GetUnit
// DESCRIPTION: Get a unit from the map.
// PARAMETERS : x - hex horizontal position
//              y - hex vertical position
// RETURNS    : the unit at the given coordinates, or NULL if no unit
//              was found there
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Unit *Map::GetUnit( short x, short y ) const {
  MapObject *o = GetObject( x, y );
  if ( o && o->IsUnit() ) return static_cast<Unit *>(o);
  return NULL;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::SetUnit
// DESCRIPTION: Put a unit on the map.
// PARAMETERS : u - unit (may be NULL)
//              x - hex horizontal position
//              y - hex vertical position
// RETURNS    : -
//
// HISTORY
//   12-06-2001 - handle dummies specially
//              - record event for turn history if building conquered
////////////////////////////////////////////////////////////////////////

void Map::SetUnit( Unit *u, short x, short y ) {
  if ( u ) {
    u->SetPosition( x, y );
    if ( u->IsGround() ) {
      if ( IsWater( x, y ) ) u->SetFlags( U_FLOATING );
      else u->UnsetFlags( U_FLOATING );
    }

    Unit *local = GetUnit( x, y );
    if ( local && local->IsTransport() ) {	// move into transport
      if ( u->IsDummy() ) u->SetFlags( U_SHELTERED );
      else static_cast<Transport *>(local)->InsertUnit( u );
    } else {
      Building *b = GetBuilding( x, y );
      if ( b ) {				// move unit into building
        if ( b->InsertUnit( u ) == 1 )	{	// building conquered
          // exchange the building entrance
          short hextype = m_data[XY2Index(x,y)], newhex;
          if ( b->Owner() ) newhex = hextype + u->Owner()->ID() - b->Owner()->ID();
          else newhex = hextype + u->Owner()->ID() - PLAYER_NONE;
          SetHexType( x, y, newhex );
          b->SetOwner( u->Owner() );

          // record event
          History *hist = Gam->GetHistory();
          if ( hist ) hist->RecordTileEvent( newhex, hextype, x, y );
        }
      } else m_objects[XY2Index(x,y)] = u;
    }
  } else m_objects[XY2Index(x,y)] = u;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::GetBuilding
// DESCRIPTION: Get a building from the map.
// PARAMETERS : x - hex horizontal position
//              y - hex vertical position
// RETURNS    : the building at the given coordinates, or NULL if no
//              building was found there
//
// HISTORY
////////////////////////////////////////////////////////////////////////

Building *Map::GetBuilding( short x, short y ) const {
  MapObject *o = GetObject( x, y );
  if ( o && o->IsBuilding() ) return static_cast<Building *>(o);
  return NULL;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::HexColor
// DESCRIPTION: Get the color of a hex for the overview map.
// PARAMETERS : xy - hex index (y * map_height + x)
// RETURNS    : hex color 
//
// HISTORY
////////////////////////////////////////////////////////////////////////

unsigned long Map::HexColor( unsigned short xy ) const {
  return( Gam->GetTerrainType(m_data[xy])->tt_color );
}

////////////////////////////////////////////////////////////////////////
// NAME       : Map::HexType
// DESCRIPTION: Get some info about a hex.
// PARAMETERS : x - hex horizontal position
//              y - hex vertical position
// RETURNS    : terrain type information
//
// HISTORY
////////////////////////////////////////////////////////////////////////

const TerrainType *Map::HexType( short x, short y ) const {
  return( Gam->GetTerrainType(m_data[XY2Index(x,y)]) );
}

