// Crimson Fields -- a game of tactical warfare
// Copyright (C) 2000-2002 Jens Granseuer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

////////////////////////////////////////////////////////////////////////
// path.cpp -- pathfinding functions
//
// History:
//  16-03-2002 - created
////////////////////////////////////////////////////////////////////////

#include "path.h"

////////////////////////////////////////////////////////////////////////
// NAME       : BasicPath::BasicPath
// DESCRIPTION: Create a new path object.
// PARAMETERS : map    - map to use for pathfinding
//              buffer - buffer to store path in; if non-NULL must be
//                       large enough to hold at least (map height x
//                       map width) bytes. If NULL, a new buffer will
//                       be allocated and freed when the path object is
//                       destroyed.
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

BasicPath::BasicPath( Map *map, char *buffer ) {
  this->map = map;
  buf_private = false;

  SetBuffer( buffer );
  Clear();
}


////////////////////////////////////////////////////////////////////////
// NAME       : BasicPath::~BasicPath
// DESCRIPTION: Destroy the Path object.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

BasicPath::~BasicPath( void ) {
  if ( buf_private ) delete [] path;
}


////////////////////////////////////////////////////////////////////////
// NAME       : BasicPath::Clear
// DESCRIPTION: Clear the path buffer.
// PARAMETERS : -
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void BasicPath::Clear( void ) {
  unsigned short size = map->MapWidth() * map->MapHeight();
  for ( int i = 0; i < size; ++i ) path[i] = -1;
}


////////////////////////////////////////////////////////////////////////
// NAME       : BasicPath::SetBuffer
// DESCRIPTION: Set a new path buffer.
// PARAMETERS : buffer - new buffer. If NULL a new buffer is allocated.
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void BasicPath::SetBuffer( char *buffer ) {
  if ( buf_private ) delete [] path;

  if ( !buffer ) {
    path = new char [map->MapWidth() * map->MapHeight()];
    buf_private = true;
  } else {
    path = buffer;
    buf_private = false;
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : BasicPath::Find
// DESCRIPTION: Search a path on the map.
// PARAMETERS : u     - unit to search a path for; this may be NULL
//                      only in some special cases (see subclasses)
//              start - hex to start from (can be different from the
//                      unit's current position)
//              end   - destination hex
//              qual  - path quality/speed trade-off (1 is best, 10 is
//                      worst but fastest); Default PATH_BEST.
//              off   - sometimes you don't want to reach the target
//                      hex proper but only get close to it. This is
//                      the maximum distance to the destination for a
//                      path to be considered valid. Default 0.
// RETURNS    : approximate number of turns to reach the destination
//              hex, or -1 if no valid path was found
//
// HISTORY
//   16-03-2002 - adapted from previous implementation
////////////////////////////////////////////////////////////////////////

short BasicPath::Find( const Unit *u, const Point &start, const Point &end,
                       unsigned char qual, unsigned char off ) {
  PathNode *pnode;
  List openlist, donelist;
  this->start = start;
  this->end = end;

  // to begin, insert the starting node in openlist
  pnode = new PathNode;
  if ( !pnode ) return -1;

  // starting from the destination would make it a bit easier to record the actual path
  // but this way we can support going towards the target until we have reached a certain
  // distance which is rather useful for AI routines...
  pnode->pos = start;
  pnode->eta = Distance( start, end ) * qual;
  pnode->cost = 0;
  openlist.AddHead( pnode );

  while( !openlist.IsEmpty() ) {
    pnode = GetBestNode( openlist );
    pnode->Remove();
    donelist.AddHead( pnode );

    // check for destination
    if ( StopSearch( pnode, off ) ) break;

    // not there yet, so let's look at the hexes around it
    for ( short dir = NORTH; dir <= NORTHWEST; ++dir ) {
      Point p;
      short cost;

      // get the hex in that direction
      if ( !map->Dir2XY( pnode->pos.x, pnode->pos.y,
                                              (Direction)dir, p ) ) {
        cost = GetCost( u, pnode->pos, p, pnode->cost );
        if ( cost >= 0 ) {
          PathNode *pnode2 = new PathNode;
          if ( !pnode2 ) return -1;

          pnode2->pos = p;
          pnode2->cost = pnode->cost + cost;
          pnode2->eta = Distance( p, end ) * qual;

          MarkBuffer( pnode2, (Direction)dir, openlist );
        }
      }
    }
  }

  // check if we reached our destination
  pnode = static_cast<PathNode *>( donelist.Head() );
  return (pnode ? RecordPath( u, pnode, off ) : -1 );
}


////////////////////////////////////////////////////////////////////////
// NAME       : BasicPath::GetStep
// DESCRIPTION: Get the path buffer content at a hex position. Usually,
//              it will contain the direction of the next step on the
//              path.
// PARAMETERS : current - hex position to look at
// RETURNS    : content of path buffer at the position of 'current'
//
// HISTORY
////////////////////////////////////////////////////////////////////////

char BasicPath::GetStep( const Point &current ) const {
  return path[map->XY2Index(current.x, current.y)];
}


// PATH - a Path is the object used for normal unit pathfinding,
//        trying to move a unit from A to B
//
//
////////////////////////////////////////////////////////////////////////
// NAME       : Path::GetBestNode
// DESCRIPTION: The best node for standard pathfinding is the one with
//              the smallest sum of cost and eta.
// PARAMETERS : open - list of open nodes
// RETURNS    : best node
//
// HISTORY
////////////////////////////////////////////////////////////////////////

PathNode *Path::GetBestNode( const List &open ) const {
  PathNode *p1, *p2;

  p1 = static_cast<PathNode *>( open.Head() );
  p2 = static_cast<PathNode *>( p1->Next() );
  while ( p2 ) {
    if ( (p2->cost + p2->eta) < (p1->cost + p1->eta) ) p1 = p2;
    p2 = static_cast<PathNode *>( p2->Next() );
  }
  return p1;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Path::StopSearch
// DESCRIPTION: For the normal pathfinder, the search ends when the
//              destination is reached (or we're sufficiently close).
// PARAMETERS : next - path node to be checked next if search is not
//                     cancelled
//              off  - minimum distance to end to abort
// RETURNS    : TRUE if search aborted, FALSE otherwise
//
// HISTORY
////////////////////////////////////////////////////////////////////////

bool Path::StopSearch( const PathNode *next, unsigned char off ) const {
  return (Distance( next->pos.x, next->pos.y, end.x, end.y ) <= off);
}

////////////////////////////////////////////////////////////////////////
// NAME       : Path::GetCost
// DESCRIPTION: Calculate the cost for the unit to move from one hex to
//              another.
// PARAMETERS : u     - unit
//              from  - source hex
//              to    - destination hex (adjacent to from)
//              sofar - cost for following the path up to this hex
// RETURNS    : cost in movement points or -1 if move is illegal
//
// HISTORY
////////////////////////////////////////////////////////////////////////

short Path::GetCost( const Unit *u, const Point &from, const Point &to,
                     short sofar ) const {
  short cost;
  const TerrainType *type = map->HexType( to.x, to.y );

  // can the unit move over that kind of terrain?
  if ( !(type->tt_type & u->Terrain()) ) cost = -1;
  else {
    Building *bld = map->GetBuilding( to.x, to.y );
    Unit *block = map->GetUnit( to.x, to.y );

    if ( bld && (!bld->Allow(u) || (to != end) ||
         u->IsSheltered()) ) cost = MCOST_UNIT;
    else if ( block && (!block->IsTransport() || (to != end) ||
         u->IsSheltered() ||
         !static_cast<Transport *>(block)->Allow(u)) ) cost = MCOST_UNIT;

    else {
      // Is the path blocked? This is the case if the unit slips through
      // between two enemy units, or one enemy unit and impassable terrain.
      // This results in movement cost equal to the unit's remaining points.
      Direction dir = XY2Dir( from.x, from.y, to.x, to.y );
      short narrow = 0, foenear = 0;
      Point p;

      if ( !map->Dir2XY( from.x, from.y, TurnLeft(dir), p ) ) {
        block = map->GetUnit( p.x, p.y );

        if ( block && (block->Owner() != u->Owner()) ) ++foenear;
        else if ( !(map->TerrainTypes(p.x,p.y) & u->Terrain()) ) ++narrow;
      } else ++narrow;

      if ( !map->Dir2XY( from.x, from.y, TurnRight(dir), p ) ) {
        block = map->GetUnit( p.x, p.y );

        if ( block && (block->Owner() != u->Owner()) ) ++foenear;
        else if ( !(map->TerrainTypes(p.x,p.y) & u->Terrain()) ) ++narrow;
      } else ++narrow;

      if ( u->IsAircraft() ) cost = MCOST_MIN;
      else cost = type->tt_move;

      if ( (foenear > 1) || ((foenear == 1) && (narrow > 0)) ) {  // blocked
        if ( sofar >= u->Moves() ) cost = u->Moves();
        else cost = MAX( u->Moves() - sofar, cost );
      }
    }
  }
  return cost;
}

////////////////////////////////////////////////////////////////////////
// NAME       : Path::MarkBuffer
// DESCRIPTION: When a hex has been checked and was found ok to move
//              onto, this hex must be properly marked in the path
//              buffer and possibly in the list of nodes, too.
// PARAMETERS : next - node which is to be added to the open list
//              dir  - direction of the step to reach next node
//              open - list of open nodes
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void Path::MarkBuffer( PathNode *next, Direction dir, List &open ) const {
  unsigned short index = map->XY2Index( next->pos.x, next->pos.y );
  PathNode *p;

  // if the new hex is not marked in path yet, mark it and put it in openlist
  if ( path[index] == -1 ) {
    path[index] = ReverseDir( dir );
    open.AddTail( next );
  } else {
    // try to find the node in openlist and replace it if new cost is smaller
    p = static_cast<PathNode *>( open.Head() );
    while ( p ) {
      if ( p->pos == next->pos ) {   // found the node. shall we replace it?
        if ( p->cost > next->cost ) {
          p->Remove();
          delete p;
          path[index] = ReverseDir( dir );
          open.AddTail( next );
        } else delete next;
        break;
      } else p = static_cast<PathNode *>( p->Next() );
    }
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : Path::RecordPath
// DESCRIPTION: Check whether a path has been found, prepare the buffer
//              and set the return code of the pathfinder.
// PARAMETERS : u    - unit searching the path
//              last - final path node
//              off  - maximum allowed distance to destination
// RETURNS    : cost of the path in turns for the unit (approximately),
//              or -1 if no valid path was found
//
// HISTORY
////////////////////////////////////////////////////////////////////////

short Path::RecordPath( const Unit *u, const PathNode *last,
                        unsigned char off ) const {
  Point p = last->pos;
  if ( Distance( p.x, p.y, end.x, end.y ) <= off ) {
    // correct the path
    // move back to beginning
    short index = map->XY2Index( p.x, p.y );
    Direction dir = (Direction)path[index], dir2;
    path[index] = -1;

    while ( p != start ) {
      map->Dir2XY( p.x, p.y, dir, p );
      index = map->XY2Index( p.x, p.y );
      dir2 = (Direction)path[index];
      path[index] = ReverseDir( dir );
      dir = dir2;
    }
    return (last->cost + u->Type()->Speed() - 1) / u->Type()->Speed();
  }
  return -1;
}



// MOVE SHADER - this is used to create the buffer for shading of
//               illegal moves when selecting a unit
//
//
////////////////////////////////////////////////////////////////////////
// NAME       : MoveShader::GetBestNode
// DESCRIPTION: For the shader the order of nodes doesn't matter. Nodes
//              with a too high cost are not added to the list, and all
//              others must be processed anyway. Therefore we always
//              use the list head.
// PARAMETERS : open - list of open nodes
// RETURNS    : best node
//
// HISTORY
////////////////////////////////////////////////////////////////////////

PathNode *MoveShader::GetBestNode( const List &open ) const {
  return static_cast<PathNode *>( open.Head() );
}

////////////////////////////////////////////////////////////////////////
// NAME       : MoveShader::StopSearch
// DESCRIPTION: The shader never stops the serach when there are nodes
//              left in the open list.
// PARAMETERS : next - path node to be checked next if search is not
//                     cancelled
//              off  - minimum distance to end to abort
// RETURNS    : TRUE if search aborted, FALSE otherwise
//
// HISTORY
////////////////////////////////////////////////////////////////////////

bool MoveShader::StopSearch( const PathNode *next, unsigned char off ) const {
  return false;
}

////////////////////////////////////////////////////////////////////////
// NAME       : MoveShader::GetCost
// DESCRIPTION: Calculate the cost for the unit to move from one hex to
//              another.
// PARAMETERS : u     - unit
//              from  - source hex
//              to    - destination hex (adjacent to from)
//              sofar - cost for following the path up to this hex
// RETURNS    : cost in movement points or -1 if move is illegal
//
// HISTORY
////////////////////////////////////////////////////////////////////////

short MoveShader::GetCost( const Unit *u, const Point &from, const Point &to,
                     short sofar ) const {
  short cost;
  if ( u->IsAircraft() ) cost = MCOST_MIN;
  else cost = map->MoveCost( to.x, to.y );

  // can the unit move over that kind of terrain?
  if ( !(map->TerrainTypes(to.x,to.y) & u->Terrain()) ) cost = -1;
  else {
    Building *bld = map->GetBuilding( to.x, to.y );
    Unit *block = map->GetUnit( to.x, to.y );

    if ( bld ) {
      if ( !u->IsSheltered() && bld->Allow(u) )
        cost = MAX( u->Moves() - sofar, cost );
      else cost = -1;

    } else if ( block ) {
      if ( block->IsTransport() && !u->IsSheltered() &&
          static_cast<Transport *>(block)->Allow(u) )
        cost = MAX( u->Moves() - sofar, cost );
      else cost = -1;

    } else {
      // Is the path blocked? This is the case if the unit slips through
      // between two enemy units, or one enemy unit and impassable terrain.
      // This results in movement cost equal to the unit's remaining points.
      Direction dir = XY2Dir( from.x, from.y, to.x, to.y );
      short narrow = 0, foenear = 0;
      Point p;

      if ( !map->Dir2XY( from.x, from.y, TurnLeft(dir), p ) ) {
        block = map->GetUnit( p.x, p.y );

        if ( block && (block->Owner() != u->Owner()) ) ++foenear;
        else if ( !(map->TerrainTypes(p.x,p.y) & u->Terrain()) ) ++narrow;
      } else ++narrow;

      if ( !map->Dir2XY( from.x, from.y, TurnRight(dir), p ) ) {
        block = map->GetUnit( p.x, p.y );

        if ( block && (block->Owner() != u->Owner()) ) ++foenear;
        else if ( !(map->TerrainTypes(p.x,p.y) & u->Terrain()) ) ++narrow;
      } else ++narrow;

      if ( (foenear > 1) || ((foenear == 1) && (narrow > 0)) ) {  // blocked
        cost = MAX( u->Moves() - sofar, cost );
      }
    }
  }

  if ( cost > u->Moves() - sofar ) cost = -1;
  return cost;
}

////////////////////////////////////////////////////////////////////////
// NAME       : MoveShader::MarkBuffer
// DESCRIPTION: When a hex has been checked and was found ok to move
//              onto, this hex must be properly marked in the path
//              buffer and possibly in the list of nodes, too.
// PARAMETERS : next - node which is to be added to the open list
//              dir  - direction of the step to reach next node
//              open - list of open nodes
// RETURNS    : -
//
// HISTORY
////////////////////////////////////////////////////////////////////////

void MoveShader::MarkBuffer( PathNode *next, Direction dir, List &open ) const {
  unsigned short index = map->XY2Index( next->pos.x, next->pos.y );
  PathNode *p;

  // if the new hex is not marked in path yet, mark it and put it in openlist
  if ( path[index] == -1 ) {
    path[index] = 1;
    open.AddTail( next );
  } else {
    // try to find the node in openlist and replace it if new cost is smaller
    p = static_cast<PathNode *>( open.Head() );
    while ( p ) {
      if ( p->pos == next->pos ) {   // found the node. shall we replace it?
        if ( p->cost > next->cost ) {
          p->Remove();
          delete p;
          open.AddTail( next );
        } else delete next;
        break;
      } else p = static_cast<PathNode *>( p->Next() );
    }
  }
}

////////////////////////////////////////////////////////////////////////
// NAME       : MoveShader::RecordPath
// DESCRIPTION: The shader doesn't have a path to prepare, but we must
//              check for valid targets which must not be shaded.
// PARAMETERS : u    - unit searching the path
//              last - final path node
//              off  - maximum allowed distance to destination
// RETURNS    : always 1
//
// HISTORY
////////////////////////////////////////////////////////////////////////

short MoveShader::RecordPath( const Unit *u, const PathNode *last,
                              unsigned char off ) const {
  Unit *target = static_cast<Unit *>( units->Head() );
  while ( target ) {
    if ( u->CanHit( target ) )
      path[map->XY2Index( target->Position().x, target->Position().y )] = 1;
    target = static_cast<Unit *>( target->Next() );
  }
  path[map->XY2Index( u->Position().x, u->Position().y )] = 1;

  return 1;
}


